package api

import (
	"net/http"
	"path"
	"strconv"
	"testing"
	"time"
)

var testOrder = Order{
	Name:        "huevos",
	Description: "huevos frescos",
	Deadline:    time.Now().Add(24 * time.Hour),
	Products: []Product{
		testProduct,
	},
}

func TestOrderAddList(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()
	tapi.addTestOrder()

	var orders []Order
	resp := tapi.do("GET", "/order", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get orders:", resp.Status)
	}

	if len(orders) != 1 {
		t.Fatal("Wrong number of orders", len(orders), orders)
	}
	if orders[0].Name != testOrder.Name {
		t.Error("Wrong name:", orders[0].Name)
	}
	if len(orders[0].Products) != 1 {
		t.Fatal("Wrong number of products", len(orders[0].Products), orders[0].Products)
	}
	if orders[0].Products[0].Price != testProduct.Price {
		t.Error("Wrong product price:", orders[0].Products[0].Price)
	}
}

func TestOrderActive(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()
	tapi.addTestOrder()

	var orders []Order
	resp := tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get orders:", resp.Status)
	}

	if len(orders) != 1 {
		t.Fatal("Wrong number of orders", len(orders), orders)
	}
	if orders[0].Name != testOrder.Name {
		t.Error("Wrong name:", orders[0].Name)
	}
}

func TestOrderPurchase(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()
	tapi.addTestOrder()

	var orders []Order
	resp := tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get orders:", resp.Status)
	}

	purchase := []OrderPurchase{
		{
			ProductCode: testProduct.Code,
			Amount:      3,
			OrderID:     orders[0].ID,
		},
	}
	resp = tapi.do("POST", "/order/purchase", purchase, nil)
	if resp.StatusCode != http.StatusCreated {
		t.Fatal("Can't create order:", resp.Status)
	}

	var transactions []Transaction
	resp = tapi.do("GET", "/transaction/mine", nil, &transactions)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(transactions) != 1 {
		t.Fatal("Wrong number of transactions", len(orders), orders)
	}
	total := 3 * testProduct.Price
	if transactions[0].Total != -total {
		t.Fatal("Wrong total", transactions[0].Total)
	}

	var member Member
	resp = tapi.do("GET", "/member/me", nil, &member)
	if resp.StatusCode != http.StatusOK {
		t.Error("Can't find the member:", resp.Status)
	}
	if member.Balance != testMember.Balance-total {
		t.Error("Wrong product balance:", member.Balance)
	}
}

func TestOrderNoDeactivation(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()

	order := testOrder
	now := time.Now()
	order.Deadline = time.Date(now.Year(), now.Month(), now.Day(), 0, 0, 0, 0, time.Local)
	resp := tapi.do("POST", "/order", order, nil)
	if resp.StatusCode != http.StatusCreated {
		t.Fatal("Can't create order:", resp.Status)
	}

	var orders []Order
	resp = tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(orders) != 1 {
		t.Fatal("Didn't find my new order")
	}

	dbPath := path.Join(tapi.testPath, "test.db")
	db, err := initDB(dbPath)
	if err != nil {
		t.Fatal("Can't initialize the db:", err)
	}
	a := api{db: db}
	a.deactivateOrders()

	resp = tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(orders) != 1 {
		t.Fatal("Didn't find my new order after deactivation")
	}
}

func TestOrderDeactivation(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()

	order := testOrder
	order.Deadline = time.Now().Add(-24 * time.Hour)
	resp := tapi.do("POST", "/order", order, nil)
	if resp.StatusCode != http.StatusCreated {
		t.Fatal("Can't create order:", resp.Status)
	}

	var orders []Order
	resp = tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(orders) != 1 {
		t.Fatal("Didn't find my new order")
	}

	purchase := []OrderPurchase{
		{
			ProductCode: testProduct.Code,
			Amount:      3,
			OrderID:     orders[0].ID,
		},
	}
	resp = tapi.doAdmin("POST", "/order/purchase", purchase, nil)
	if resp.StatusCode != http.StatusCreated {
		t.Fatal("Can't create order:", resp.Status)
	}

	dbPath := path.Join(tapi.testPath, "test.db")
	db, err := initDB(dbPath)
	if err != nil {
		t.Fatal("Can't initialize the db:", err)
	}
	a := api{db: db}
	a.deactivateOrders()

	resp = tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(orders) != 0 {
		t.Fatal("I found some orders")
	}

	total := 3 * testProduct.Price

	var transactions []Transaction
	resp = tapi.do("GET", "/transaction/mine", nil, &transactions)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get transactions:", resp.Status)
	}
	if len(transactions) != 1 {
		t.Fatal("Wrong number of transactions", len(orders), orders)
	}
	if transactions[0].Type != "refund" {
		t.Fatal("Should be a refund", transactions[0].Type)
	}
	if transactions[0].Total != total {
		t.Fatal("Wrong total:", transactions[0].Total)
	}

	var member Member
	resp = tapi.do("GET", "/member/me", nil, &member)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't member:", resp.Status)
	}
	if member.Balance != testMember.Balance+total {
		t.Fatal("Wrong member balance:", member.Balance, testMember.Balance, total)
	}
}

func TestGetOrder(t *testing.T) {
	tapi := newTestAPI(t)
	defer tapi.close()
	tapi.addTestMember()
	tapi.addTestOrder()

	var orders []Order
	resp := tapi.do("GET", "/order/active", nil, &orders)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get orders:", resp.Status)
	}
	if len(orders) != 1 {
		t.Fatal("Didn't find my new order")
	}
	id := strconv.Itoa(int(orders[0].ID))

	var body OrderGetResponse
	resp = tapi.do("GET", "/order/"+id, nil, &body)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get order:", resp.Status)
	}
	if body.Transaction != nil {
		t.Error("Unexpected transaction", body.Transaction)
	}
	if body.Order.Name != testOrder.Name {
		t.Error("Wrong name:", body.Order.Name)
	}

	purchase := []OrderPurchase{
		{
			ProductCode: testProduct.Code,
			Amount:      3,
			OrderID:     orders[0].ID,
		},
	}
	resp = tapi.do("POST", "/order/purchase", purchase, nil)
	if resp.StatusCode != http.StatusCreated {
		t.Fatal("Can't create order:", resp.Status)
	}

	resp = tapi.do("GET", "/order/"+id, nil, &body)
	if resp.StatusCode != http.StatusOK {
		t.Fatal("Can't get order:", resp.Status)
	}
	if body.Transaction == nil {
		t.Fatal("There is no transaction")
	}
	if body.Transaction.Type != "order" {
		t.Error("Wrong transaction type", body.Transaction.Type)
	}
}

func (tapi *testAPI) addTestOrder() {
	resp := tapi.do("POST", "/order", testOrder, nil)
	if resp.StatusCode != http.StatusCreated {
		tapi.t.Fatal("Can't create order:", resp.Status)
	}
}
